import threading
from datetime import datetime
from typing import List

from retrying import retry

from sidecar.azure.azure_cosmos_service import AzureCosmosService
from sidecar.azure.retrying_helpers import retry_if_connection_error, retry_if_result_none_or_empty
from sidecar.const import Const


class AzureStatusMaintainer:
    ROOT_SPEC_KEY = "spec"
    EXPECTED_APPS_KEY = "expected_apps"
    ROOT_APPS_KEY = "apps"
    INSTANCES_KEY = "instances"

    def __init__(self, cosmos_service: AzureCosmosService, sandbox_id: str):
        self._cosmos_service = cosmos_service
        self._sandbox_id = sandbox_id
        self._lock = threading.RLock()

        self._cached_realtime_apps = None
        self._cached_spec = None
        self._cache_sandbox_data()

    def _cache_sandbox_data(self):
        sandbox_document = self._get_sandbox_document_with_retries()
        if not sandbox_document:
            raise Exception("Failed to get sandbox data from cosmos_db after 5 retries")

        self._cached_realtime_apps = sandbox_document[self.ROOT_APPS_KEY]
        self._cached_spec = sandbox_document[self.ROOT_SPEC_KEY]

    @retry(stop_max_attempt_number=5, wait_fixed=1000, retry_on_result=retry_if_result_none_or_empty,
           retry_on_exception=retry_if_connection_error)
    def _get_sandbox_document_with_retries(self):
        return self._cosmos_service.find_document_by_id(id=self._sandbox_id)

    def _construct_instances_key_path(self, instance_logical_id):
        return '{apps_key}.{instance_logical_id}.{instances_key}'.format(apps_key=self.ROOT_APPS_KEY,
                                                                         instance_logical_id=instance_logical_id,
                                                                         instances_key=self.INSTANCES_KEY)

    def update_app_instance_status(self, instance_logical_id, instance_id, app_name, status):
        with self._lock:

            instances_under_logical_id = self._cached_realtime_apps[instance_logical_id][self.INSTANCES_KEY]
            apps_under_instance_id = instances_under_logical_id.setdefault(instance_id, {"apps": {}})["apps"]
            apps_under_instance_id[app_name] = {Const.APP_STATUS_TAG: status}

            instances_key_path = self._construct_instances_key_path(instance_logical_id)
            self._cosmos_service.update_document(id=self._sandbox_id,
                                                 update_param={
                                                     '$set': {
                                                         instances_key_path: instances_under_logical_id
                                                     }})

    def get_app_names_on_instance(self, instance_logical_id: str) -> List[str]:
        return self._cached_spec[self.EXPECTED_APPS_KEY][instance_logical_id]["apps"]

    def update_sandbox_end_status(self, sandbox_deployment_end_status: str):
        # assume that DB-level lock is good enough here
        self._cosmos_service.update_document(id=self._sandbox_id,
                                             update_param={
                                                 '$set': {
                                                     Const.SANDBOX_DEPLOYMENT_END_STATUS: sandbox_deployment_end_status,
                                                     Const.SANDBOX_DEPLOYMENT_END_STATUS_v2: sandbox_deployment_end_status
                                                 }})

    def update_sandbox_start_status(self, sandbox_start_time: datetime):
        # assume that DB-level lock is good enough here
        self._cosmos_service.update_document(id=self._sandbox_id,
                                             update_param={
                                                 '$set': {
                                                     Const.SANDBOX_START_TIME: str(sandbox_start_time),
                                                     Const.SANDBOX_START_TIME_v2: str(sandbox_start_time)
                                                 }})



from sidecar.azure.azure_status_maintainer import AzureStatusMaintainer
from sidecar.sandbox_deployment_end_updater import ISandboxDeploymentEndUpdater


class AzureSandboxDeploymentEndUpdater(ISandboxDeploymentEndUpdater):
    def __init__(self, status_maintainer: AzureStatusMaintainer):
        super(AzureSandboxDeploymentEndUpdater, self).__init__()
        self._status_maintainer = status_maintainer

    def _set_deployment_end_details(self, deployment_end_status: str):
        self._status_maintainer.update_sandbox_end_status(deployment_end_status)



import os
import signal
import subprocess
from datetime import datetime
from logging import Logger
from typing import List, Dict

from sidecar.app_instance_identifier import IIdentifier
from sidecar.health_check.health_check_executor_logger import HealthCheckExecutorLogger
from sidecar.non_blocking_stream_reader import NonBlockingStreamReader
from sidecar.utils import CallsLogger


class HealthCheckExecutor:

    def __init__(self,
                 env_vars: Dict[str, str],
                 app_timeouts: Dict[str, int],
                 executor_logger: HealthCheckExecutorLogger,
                 logger: Logger):
        self._executor_logger = executor_logger
        self._app_timeouts = app_timeouts
        self._env_vars = env_vars
        self._logger = logger

    @CallsLogger
    def start(self, identifier: IIdentifier, cmd: List[str]) -> bool:
        timeout = self._app_timeouts.get(identifier.name)
        self._executor_logger.log_start(identifier=identifier, cmd=cmd, timeout=timeout)

        env = self._get_env_vars(identifier)

        start = datetime.now()
        timed_out = False
        read_interval = 0.5

        # run healthcheck command in subprocess and redirect its outputs to subprocess' stdout
        # read stdout line by line until subprocess ended or until timeout and send it to cloud logger
        # if timeout occurred kill healthcheck subprocess
        with subprocess.Popen(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              shell=True,
                              preexec_fn=os.setsid,
                              universal_newlines=True,
                              env=env) as p:
            try:
                stdout_stream_reader = NonBlockingStreamReader(stream=p.stdout, interval=read_interval)
                stderr_stream_reader = NonBlockingStreamReader(stream=p.stderr, interval=read_interval)
                self._logger.info('running command {0}'.format(cmd))

                while True:
                    line = stdout_stream_reader.read_line(read_interval)
                    if line:
                        self._executor_logger.log_line(line, identifier)

                    line = stderr_stream_reader.read_line(read_interval)
                    if line:
                        self._executor_logger.log_line(line, identifier, True)

                    elapsed = datetime.now() - start

                    if elapsed.total_seconds() > timeout:
                        stdout_stream_reader.stop()
                        stderr_stream_reader.stop()
                        raise subprocess.TimeoutExpired(cmd=cmd, timeout=timeout)

                    if p.poll() is not None:
                        while True:
                            line = stdout_stream_reader.read_line(read_interval)
                            if line:
                                self._executor_logger.log_line(line, identifier)
                            else:
                                stdout_stream_reader.stop()
                                break

                        while True:
                            line = stderr_stream_reader.read_line(read_interval)
                            if line:
                                self._executor_logger.log_line(line, identifier, True)
                            else:
                                stderr_stream_reader.stop()
                                break
                        break
            except subprocess.TimeoutExpired as ex:
                self._executor_logger.log_timeout(timeout=ex.timeout, identifier=identifier)
                self._kill_process(identifier=identifier, process=p)
                timed_out = True
            finally:
                if timed_out:
                    return False
                process_exit_code = p.returncode
                if process_exit_code == 0:
                    self._executor_logger.log_success(identifier=identifier)
                    return True
                else:
                    self._executor_logger.log_error(identifier=identifier, exit_code=process_exit_code)
                    return False

    @CallsLogger
    def _kill_process(self, identifier, process):
        try:
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
        except ProcessLookupError as ex:
            self._logger.exception('Could not kill process, pid {} due to {}'.format(
                process.pid,
                str(ex)))

    @CallsLogger
    def _get_env_vars(self, identifier):
        # get process environment variables and mix them with user defined application specific environment
        # variables
        app_env = self._env_vars.get(identifier.name)
        return {**os.environ, **({} if app_env is None else app_env)}



import inspect
import logging
import os
import sys

import time

import datetime
from logging import Logger

from sidecar.const import Const


class Utils:

    @staticmethod
    def stop_on_debug():
        while not sys.gettrace():
            time.sleep(0.5)

    @staticmethod
    def read_log(app_name: str) -> str:
        file_path = Const.get_app_log_file(app_name)
        with open(file_path, 'r') as application_log:
            return application_log.read()

    @staticmethod
    def get_timestamp():
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def wait_for(func, interval_sec: int = 10, max_retries: int = 10, error: str = "", silent: bool = False):
        current_retry = 0
        while current_retry < max_retries:
            try:
                if func():
                    return
                else:
                    if not silent:
                        print('re-try {} out of {}'.format(current_retry, max_retries))
                    time.sleep(interval_sec)
                    current_retry += 1
                    if current_retry >= max_retries:
                        raise Exception('max retries for wait_for is exhausted with message: {}'.format(error))
            except Exception as e:
                raise Exception('wait_for function exited due to an exception: {}'.format(e))

    @staticmethod
    def retry_on_exception(func,
                           logger: Logger = None,
                           logger_msg: str = "",
                           interval: int = 1,
                           timeout: int = 10):
        remains = timeout / interval
        while True:
            try:
                return func()
            except Exception as e:
                if remains == 0:
                    raise e
                else:
                    if logger:
                        logger.info("{} ({} remaining attempts)".format(logger_msg, remains))
                remains -= 1
                time.sleep(interval)


class CallsLogger:
    logger = None

    @classmethod
    def set_logger(cls, logger):
        cls.logger = logger

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        bound_args = inspect.signature(self.func).bind(*args, **kwargs)
        bound_args.apply_defaults()

        args_list = ["{}: '{}'".format(k, v) for (k, v) in bound_args.arguments.items()]
        if len(bound_args.arguments) > 0 and bound_args.arguments.get('self'):
            args_list = args_list[1:]

        self.logger.info(self.func.__qualname__ + "(" + ", ".join(args_list) + ")")

        return self.func(**bound_args.arguments)



from typing import List


class AppRequestInfo:
    def __init__(self, name: str, instances_count: int, app_dependencies: List[str] = None):
        self.name = name
        self.instances_count = instances_count
        self.app_dependencies = app_dependencies if app_dependencies is not None else []
